import numpy as np
import dgl
import torch.nn as nn

def build_graph(data, train_test_ratio,embed_size):
    data['item'] = data['item'].astype(str)
    words = data['user'].unique().tolist() + data['item'].unique().tolist()
    le = preprocessing.LabelEncoder()
    le.fit(words)
    src = np.array(le.transform(data.user))
    dst = np.array(le.transform(data.item))
    # Edges are directional in DGL; Make them bi-directional.
    u = np.concatenate([src, dst])
    v = np.concatenate([dst, src])
    # Construct a DGLGraph
    g = dgl.DGLGraph((u, v))
    print(g.number_of_nodes())
    print(len(g.nodes()))
    print(len(data.user.unique().tolist()+data.item.unique().tolist()))
    dict_node = {raw:node for raw,node in zip(data.user.unique().tolist()+data.item.unique().tolist(),g.nodes())}
    features = nn.Embedding(g.number_of_nodes(), embed_size)
    labels = th.LongTensor(g.nodes())
    train_mask = th.BoolTensor([True]* int(g.number_of_nodes()*train_test_ratio) +
                               [False]*(len(g.nodes())-int(g.number_of_nodes()*train_test_ratio)))
    test_mask = th.BoolTensor([False]* int(g.number_of_nodes()*train_test_ratio) +
                               [True]*(len(g.nodes())-int(g.number_of_nodes()*train_test_ratio)))
    return g, features, labels, train_mask, test_mask



import dgl
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from models.utils import *
from dgl.nn.pytorch import GraphConv
import time

class GCNLayer(nn.Module):
    def __init__(self, in_feats, out_feats):
        super(GCNLayer, self).__init__()
        self.linear = nn.Linear(in_feats, out_feats)

    def forward(self, g, feature):
        # Creating a local scope so that all the stored ndata and edata
        # (such as the `'h'` ndata below) are automatically popped out
        # when the scope exits.
        with g.local_scope():
            g.ndata['h'] = feature
            g.update_all(gcn_msg, gcn_reduce)
            h = g.ndata['h']
            return self.linear(h)

class Net(nn.Module):
    def __init__(self,in_feats, out_feats):
        super(Net, self).__init__()
        self.layer1 = GCNLayer(in_feats, out_feats)
        self.layer2 = GCNLayer(out_feats, 32)

    def forward(self, g, features):
        x = F.relu(self.layer1(g, features))
        x = self.layer2(g, x)
        return x

class GCN(object):
    def __init__(self, data, embed_size = 128, task = 'NodeClassification',train_test_ratio = 0.8):
        self.g,
        self.features,
        self.labels,
        self.train_mask,
        self.test_mask = build_graph(data, train_test_ratio,embed_size=128)

    def evaluate(self,model, g, features, labels, mask):
        model.eval()
        with th.no_grad():
            logits = model(g, features)
            logits = logits[mask]
            labels = labels[mask]
            _, indices = th.max(logits, dim=1)
            correct = th.sum(indices == labels)
            return correct.item() * 1.0 / len(labels)

    def train(self,verbose = 1,epochs = 10):
        net = Net(self.g.number_of_nodes(), self.embed_size)
        optimizer = th.optim.Adam(net.parameters(), lr=1e-2)
        dur = []
        for epoch in range(epochs):
            if epoch >=3:
                t0 = time.time()

            net.train()
            logits = net(self.g, self.features)
            logp = F.log_softmax(logits, 1)
            loss = F.nll_loss(logp[self.train_mask], labels[self.train_mask])

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if epoch >=3:
                dur.append(time.time() - t0)

            acc = evaluate(net, self.g, self.features, self.labels, self.test_mask)
            print("Epoch {:05d} | Loss {:.4f} | Test Acc {:.4f} | Time(s) {:.4f}".format(
                    epoch, loss.item(), acc, np.mean(dur)))



'''
input data:
a dataframe with two columns: user item

output:
a embedding lookup dictionary {'user_id/item_id':[vector]}

'''
from gensim.models import Word2Vec
#import deepneighbor.config as config
from deepneighbor.utils import generate_sentences, generate_sentences_dw
from annoy import AnnoyIndex
from sklearn import preprocessing
from deepneighbor.models import *


EMBED_SIZE = 128
WINDOW_SIZE = 10
ITER = 5
WORKERS = 3


class Embed(object):
    def __init__(self,data,model='w2v'):
        '''
        data: a dataframe: user, item

        model: 'w2v','deepwalk','gcn'
        '''

        self.data = data
        self.w2v_model = None
        self._annoy = None
        self._le = None
        self._embeddings = {}
        self.model_type = model

        if self.model_type == 'w2v':
            if 'time' in data.columns.tolist():
                self.sentences = generate_sentences_dw(data)
            else:
                self.sentences = generate_sentences(data)
        if self.model_type == 'deepwalk':
            pass
        if self.model_type == 'gcn':
            pass

        self.dimension = 0




    def train(self, embed_size=128, window_size=5, workers=3, iter=5, **kwargs):

        if self.model_type == 'w2v':
            kwargs["sentences"] = self.sentences
            kwargs["min_count"] = kwargs.get("min_count", 0)
            kwargs["size"] = embed_size
            kwargs["sg"] = 1  # skip gram
            kwargs["hs"] = 1  # deepwalk use Hierarchical Softmax
            kwargs["workers"] = workers
            kwargs["window"] = window_size
            kwargs["iter"] = iter
            self.dimension = embed_size

            print(f"There are {self.data.user.nunique()} users")
            print(f"There are {self.data.item.nunique()} items")

            print("Learning embedding vectors...")
            model_w2v = Word2Vec(**kwargs)
            print("Learning embedding vectors done!")

            self.w2v_model = model_w2v

            self._annoy = AnnoyIndex(self.dimension, 'angular')

            words = self.data['user'].unique().tolist() + self.data['item'].unique().tolist()
            self._le = preprocessing.LabelEncoder()
            self._le.fit(words)
            for word in words:
                self._annoy.add_item(self._le.transform([word])[0],self.w2v_model.wv[word])

            self._annoy.build(-1)

        if self.model_type == 'gcn':
            model = gcn.GCN(data, embed_size = 128, task = 'NodeClassification',train_test_ratio = 0.8,embed_size=128)
            model.train(verbose = 1,epochs = 10)
            

        #return model_w2v


    # def get_embeddings(self,):
    #     if self.w2v_model is None:
    #         print("model not train")
    #         return {}
    #
    #     self._embeddings = {}
    #     words = self.data['user'].unique().tolist() + self.data['item'].unique().tolist()
    #     for word in words:
    #         self._embeddings[word] = self.w2v_model.wv[word]
    #
    #     return self._embeddings

    def search(self, seed,k = 5):
        '''
        seed: seed item to find nearest neighbor
        k: number of cloest neighhbors
        '''

        a_return = self._annoy.get_nns_by_item(self._le.transform([seed])[0], k)
        return list(self._le.inverse_transform(a_return))
